@@ -24,7 +24,7 @@
 
 setClassUnion("NULL_OR_list", c("NULL", "list"))
 
-.SVT_VERSION <- 0L
+SVT_VERSION <- 1L
 
 setClass("SVT_SparseArray",
     contains="SparseArray",
@@ -35,7 +35,7 @@ setClass("SVT_SparseArray",
     ),
     prototype(
         type="logical",
-        .svt_version=.SVT_VERSION
+        .svt_version=SVT_VERSION
     )
 )
 
@@ -50,22 +50,39 @@ setClass("SVT_SparseMatrix",
 ### Not exported (for internal use only).
 svt_version <- function(x)
 {
-    stopifnot(is(x, "SVT_SparseArray"))
+    stopifnot(is(x, "SVT_SparseArray") || is(x, "NaArray"))
     if (.hasSlot(x, ".svt_version")) x@.svt_version else 0L
 }
 
 check_svt_version <- function(x)
 {
-    if (svt_version(x) != 0L && !is.null(x@SVT)) {
-        pkg_version <- as.character(packageVersion("SparseArray"))
-        stop(wmsg("Incompatible SVT_SparseArray object detected: object uses ",
-                  "version ", svt_version(x), " of the SVT internal layout ",
-                  "which is not compatible with versions < 1.5.0 of the ",
-                  "SparseArray package (your version is ", pkg_version, ")."))
+    if (svt_version(x) != 0L ||
+        is(x, "SVT_SparseArray") && is.null(x@SVT) ||
+        is(x, "NaArray") && is.null(x@NaSVT))
+    {
+        return(invisible(NULL))
     }
+    pkg_version <- as.character(packageVersion("SparseArray"))
+    stop(wmsg("Old ", class(x)[[1L]], " object detected: object uses ",
+              "version 0 of the SVT internal layout which is not ",
+              "compatible with versions >= 1.5.0 of the SparseArray ",
+              "package (your version is ", pkg_version, ")."))
 }
 
 
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Validity
+###
+
+.validate_SVT_SparseArray <- function(x)
+{
+    if (!isSingleString(x@type))
+        return("'type' slot must be a single string")
+    TRUE
+}
+setValidity2("SVT_SparseArray", .validate_SVT_SparseArray)
+
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Going back and forth between SVT_SparseArray and SVT_SparseMatrix
 ###
@@ -107,11 +124,6 @@ setMethod("coerce", c("SVT_SparseMatrix", "SparseArray"),
 )
 
 
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Validity
-###
-
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### type() getter and setter
 ###
@@ -128,8 +140,8 @@ setMethod("type", "SVT_SparseArray", function(x) x@type)
     if (value == x_type)
         return(x)
 
-    new_SVT <- SparseArray.Call("C_set_SVT_SparseArray_type",
-                                x@dim, x@type, x@SVT, value)
+    new_SVT <- SparseArray.Call("C_set_SVT_type",
+                                x@dim, x@type, x@SVT, FALSE, value)
     BiocGenerics:::replaceSlots(x, type=value, SVT=new_SVT, check=FALSE)
 }
 
@@ -137,32 +149,46 @@ setReplaceMethod("type", "SVT_SparseArray", .set_SVT_SparseArray_type)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### nzcount() and nzwhich()
+### is_nonzero(), nzcount(), nzwhich(), nzvals(), `nzvals<-`()
 ###
 
-### Note that like for the length of atomic vectors in base R, the "nonzero
-### count" will be returned as a double if it's > .Machine$integer.max
-.get_SVT_SparseArray_nzcount <- function(x)
+### Returns a "logical" SVT_SparseArray object.
+.is_nonzero_SVT <- function(x)
 {
     stopifnot(is(x, "SVT_SparseArray"))
     check_svt_version(x)
-    SparseArray.Call("C_nzcount_SVT_SparseArray", x@dim, x@SVT)
+    new_SVT <- SparseArray.Call("C_is_nonzero_SVT", x@dim, x@SVT)
+    BiocGenerics:::replaceSlots(x, type="logical", SVT=new_SVT, check=FALSE)
 }
 
-setMethod("nzcount", "SVT_SparseArray", .get_SVT_SparseArray_nzcount)
+setMethod("is_nonzero", "SVT_SparseArray", .is_nonzero_SVT)
+
+### Note that like for the length of atomic vectors in base R, the "nonzero
+### count" will be returned as a double if it's > .Machine$integer.max
+.nzcount_SVT <- function(x)
+{
+    stopifnot(is(x, "SVT_SparseArray"))
+    check_svt_version(x)
+    SparseArray.Call("C_nzcount_SVT", x@dim, x@SVT)
+}
+
+setMethod("nzcount", "SVT_SparseArray", .nzcount_SVT)
 
 ### Returns an integer vector of length nzcount(x) if 'arr.ind=FALSE', or
 ### a matrix with nzcount(x) rows if 'arr.ind=TRUE'.
-.nzwhich_SVT_SparseArray <- function(x, arr.ind=FALSE)
+.nzwhich_SVT <- function(x, arr.ind=FALSE)
 {
     stopifnot(is(x, "SVT_SparseArray"))
     check_svt_version(x)
     if (!isTRUEorFALSE(arr.ind))
         stop(wmsg("'arr.ind' must be TRUE or FALSE"))
-    SparseArray.Call("C_nzwhich_SVT_SparseArray", x@dim, x@SVT, arr.ind)
+    SparseArray.Call("C_nzwhich_SVT", x@dim, x@SVT, arr.ind)
 }
 
-setMethod("nzwhich", "SVT_SparseArray", .nzwhich_SVT_SparseArray)
+setMethod("nzwhich", "SVT_SparseArray", .nzwhich_SVT)
+
+### TODO: Implement optimized nzvals() and `nzvals<-`() methods for
+### SVT_SparseArray objects.
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -193,32 +219,64 @@ new_SVT_SparseArray <- function(dim, dimnames=NULL,
     stopifnot(is(from, "SVT_SparseArray"))
     check_svt_version(from)
     SparseArray.Call("C_from_SVT_SparseArray_to_Rarray",
-                     from@dim, dimnames(from), from@type, from@SVT)
+                     from@dim, dimnames(from), from@type, from@SVT, FALSE)
 }
 
 ### S3/S4 combo for as.array.SVT_SparseArray
 as.array.SVT_SparseArray <- function(x, ...) .from_SVT_SparseArray_to_array(x)
 setMethod("as.array", "SVT_SparseArray", as.array.SVT_SparseArray)
 
-.build_SVT_SparseArray_from_array <- function(x, type=NA)
+.build_SVT_SparseArray_from_array <- function(x, dimnames=NULL, type=NA)
 {
     stopifnot(is.array(x))
+    if (is.null(dimnames)) {
+        ans_dimnames <- dimnames(x)
+    } else {
+        ans_dimnames <- S4Arrays:::normarg_dimnames(dimnames, dim(x))
+    }
     if (identical(type, NA))
         type <- type(x)
-    ans_SVT <- SparseArray.Call("C_build_SVT_from_Rarray", x, type)
-    new_SVT_SparseArray(dim(x), dimnames(x), type, ans_SVT, check=FALSE)
+    ans_SVT <- SparseArray.Call("C_build_SVT_from_Rarray", x, type, FALSE)
+    new_SVT_SparseArray(dim(x), ans_dimnames, type, ans_SVT, check=FALSE)
 }
 
 setAs("array", "SVT_SparseArray",
     function(from) .build_SVT_SparseArray_from_array(from)
 )
+setAs("array", "SparseArray",
+    function(from) .build_SVT_SparseArray_from_array(from)
+)
 setAs("matrix", "SVT_SparseMatrix",
     function(from) .build_SVT_SparseArray_from_array(from)
 )
+setAs("matrix", "SparseMatrix",
+    function(from) .build_SVT_SparseArray_from_array(from)
+)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Going back and forth between SVT_SparseMatrix and *gCMatrix objects
+### Make an SVT_SparseMatrix object from CSC components
+###
+
+### NOT exported but used in the HDF5Array package!
+### 'row_indices' must be an integer vector containing 0-based or 1-based
+### row indices. Note that the indices in 'row_indices' are not required
+### to be increasing within columns. However, 'row_indices' should NOT
+### contain duplicates within columns. This is NOT checked!
+make_SVT_SparseMatrix_from_CSC <- function(dim, indptr, data, row_indices,
+                                           indices.are.1based=FALSE,
+                                           dimnames=NULL)
+{
+    stopifnot(is.integer(row_indices), isTRUEorFALSE(indices.are.1based))
+    ans_SVT <- SparseArray.Call("C_build_SVT_from_CSC",
+                                dim, indptr, data, row_indices,
+                                indices.are.1based)
+    new_SVT_SparseArray(dim, dimnames, type(data), ans_SVT, check=FALSE)
+}
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Going back and forth between SVT_SparseMatrix and CsparseMatrix
 ###
 
 .make_CsparseMatrix_from_SVT_SparseMatrix <- function(from, to_type=NULL)
@@ -264,20 +322,70 @@ setAs("SVT_SparseMatrix", "dgCMatrix", .from_SVT_SparseMatrix_to_dgCMatrix)
 setAs("SVT_SparseMatrix", "lgCMatrix", .from_SVT_SparseMatrix_to_lgCMatrix)
 setAs("SVT_SparseMatrix", "ngCMatrix", .from_SVT_SparseMatrix_to_ngCMatrix)
 
-.build_SVT_SparseMatrix_from_CsparseMatrix <- function(x, type=NA)
+.build_SVT_SparseMatrix_from_CsparseMatrix <- function(x, dimnames=NULL,
+                                                          type=NA)
 {
     stopifnot(is(x, "CsparseMatrix"))
+
+    ## Turn any [d|l|n]gCMatrix derivative (e.g. TestColMatrix object 'y'
+    ## defined in alabaster.matrix/tests/testthat/test-SparseMatrix.R) into
+    ## a [d|l|n]gCMatrix **instance**. We should not need to do this. Only
+    ## reason we do it is because we don't know how to test for inheritance
+    ## at the C level (Rf_inherits() doesn't seem to work properly on S4
+    ## objects). More precisely, without this coercion, C function
+    ## get_gCMatrix_subtype() defined in src/SVT_SparseArray_class.c won't
+    ## be able to recognize a [d|l|n]gCMatrix derivative so will reject it.
+    x <- as(x, "CsparseMatrix")
+
+    if (is.null(dimnames)) {
+        ans_dimnames <- dimnames(x)
+    } else {
+        ans_dimnames <- S4Arrays:::normarg_dimnames(dimnames, dim(x))
+    }
     if (identical(type, NA))
         type <- type(x)
     ans_SVT <- SparseArray.Call("C_build_SVT_from_CsparseMatrix", x, type)
-    new_SVT_SparseArray(dim(x), dimnames(x), type, ans_SVT, check=FALSE)
+    new_SVT_SparseArray(dim(x), ans_dimnames, type, ans_SVT, check=FALSE)
 }
 
 setAs("CsparseMatrix", "SVT_SparseMatrix",
     function(from) .build_SVT_SparseMatrix_from_CsparseMatrix(from)
 )
 
-setAs("Matrix", "SVT_SparseArray", function(from) as(from, "SVT_SparseMatrix"))
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Going back and forth between SVT_SparseMatrix and TsparseMatrix
+###
+
+setAs("SVT_SparseMatrix", "dgTMatrix",
+    function(from) as(as(from, "dgCMatrix"), "TsparseMatrix")
+)
+setAs("SVT_SparseMatrix", "lgTMatrix",
+    function(from) as(as(from, "lgCMatrix"), "TsparseMatrix")
+)
+setAs("SVT_SparseMatrix", "ngTMatrix",
+    function(from) as(as(from, "ngCMatrix"), "TsparseMatrix")
+)
+
+setAs("TsparseMatrix", "SVT_SparseMatrix",
+    function(from) as(as(from, "CsparseMatrix"), "SVT_SparseMatrix")
+)
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Coercion from a Matrix derivative to SparseMatrix or SparseArray
+###
+
+### Coercing a sparseMatrix derivative (e.g. a CsparseMatrix or TsparseMatrix
+### derivative) to SparseMatrix produces an SVT_SparseMatrix object.
+### RsparseMatrix is the exception (see COO_SparseArray-class.R).
+setAs("sparseMatrix", "SparseMatrix",
+    function(from) as(from, "SVT_SparseMatrix")
+)
+
+setAs("Matrix", "SparseArray",
+    function(from) as(from, "SparseMatrix")
+)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -304,9 +412,15 @@ setAs("SVT_SparseMatrix", "COO_SparseMatrix",
     .from_SVT_SparseArray_to_COO_SparseArray
 )
 
-.build_SVT_SparseArray_from_COO_SparseArray <- function(x, type=NA)
+.build_SVT_SparseArray_from_COO_SparseArray <- function(x, dimnames=NULL,
+                                                           type=NA)
 {
     stopifnot(is(x, "COO_SparseArray"))
+    if (is.null(dimnames)) {
+        ans_dimnames <- dimnames(x)
+    } else {
+        ans_dimnames <- S4Arrays:::normarg_dimnames(dimnames, dim(x))
+    }
     if (identical(type, NA)) {
         type <- type(x)
     } else {
@@ -318,7 +432,7 @@ setAs("SVT_SparseMatrix", "COO_SparseMatrix",
     }
     ## We start with an allzero SVT_SparseArray object and subassign
     ## the nonzero data to it.
-    ans <- new_SVT_SparseArray(x@dim, x@dimnames, type, check=FALSE)
+    ans <- new_SVT_SparseArray(x@dim, ans_dimnames, type, check=FALSE)
     ans[x@nzcoo] <- x@nzdata
     ans
 }
@@ -332,7 +446,55 @@ setAs("COO_SparseMatrix", "SVT_SparseMatrix",
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### SVT_SparseArray() constructor
+### Default coercions to SparseArray or SVT_SparseArray
+###
+### Given a DelayedArray object or any out-of-memory array-like object 'x':
+### - as.array(x) is the standard way to realize it as an ordinary array, that
+###   is, as a **dense** array);
+### - as(x, "SparseArray") is the standard way to realize it in memory as a
+###   SparseArray derivative (SVT_SparseArray or COO_SparseArray object), that
+###   is as a **sparse** array;
+### - as(x, "SVT_SparseArray") is the standard way to realize it in memory
+###   as an SVT_SparseArray object.
+###
+
+### Similar to as.array() method for Array object (defined in the S4Arrays
+### package) but based on extract_sparse_array() instead of extract_array(),
+### and without the 'drop' argument.
+### Returns an SVT_SparseArray or COO_SparseArray object.
+.as_SparseArray <- function(x)
+{
+    if (!is_sparse(x)) {
+        ## Go thru .dense2sparse().
+        return(as(x, "COO_SparseArray"))
+    }
+    index <- vector("list", length=length(dim(x)))
+    ans <- extract_sparse_array(x, index)
+    S4Arrays:::set_dimnames(ans, dimnames(x))
+}
+setAs("ANY", "SparseArray", function(from) .as_SparseArray(from))
+
+.as_SparseMatrix <- function(x)
+{
+    x_ndim <- length(dim(x))
+    if (x_ndim != 2L)
+        stop(wmsg("cannot coerce ", class(x)[[1L]], " object ",
+                  "with ", x_ndim, " dimensions to SparseMatrix ",
+                  "(object to coerce must have 2 dimensions)"))
+    .as_SparseArray(x)
+}
+setAs("ANY", "SparseMatrix", function(from) .as_SparseMatrix(from))
+
+setAs("ANY", "SVT_SparseArray",
+    function(from) as(as(from, "SparseArray"), "SVT_SparseArray")
+)
+setAs("ANY", "SVT_SparseMatrix",
+    function(from) as(as(from, "SparseMatrix"), "SVT_SparseMatrix")
+)
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### The SVT_SparseArray() constructor
 ###
 
 .new_empty_SVT_SparseArray <- function(type=NA)
@@ -342,53 +504,55 @@ setAs("COO_SparseMatrix", "SVT_SparseMatrix",
     new2("SVT_SparseArray", type=type, check=FALSE)
 }
 
-.SVT_SparseArray <- function(x, type=NA)
+.SVT_SparseArray <- function(x, dimnames=NULL, type=NA)
 {
     if (is.array(x))
-        return(.build_SVT_SparseArray_from_array(x, type=type))
+        return(.build_SVT_SparseArray_from_array(x,
+                                      dimnames=dimnames, type=type))
 
     if (is(x, "CsparseMatrix"))
-        return(.build_SVT_SparseMatrix_from_CsparseMatrix(x, type=type))
+        return(.build_SVT_SparseMatrix_from_CsparseMatrix(x,
+                                      dimnames=dimnames, type=type))
 
     if (is(x, "COO_SparseArray"))
-        return(.build_SVT_SparseArray_from_COO_SparseArray(x, type=type))
+        return(.build_SVT_SparseArray_from_COO_SparseArray(x,
+                                      dimnames=dimnames, type=type))
 
     ans <- as(x, "SVT_SparseArray")
+    ans <- S4Arrays:::set_dimnames(ans, dimnames)
     if (!identical(type, NA))
         type(ans) <- type
     ans
 }
 
-SVT_SparseArray <- function(x, type=NA)
+SVT_SparseArray <- function(x, dim=NULL, dimnames=NULL, type=NA)
 {
     if (!identical(type, NA))
         type <- S4Arrays:::normarg_array_type(type, "the requested type")
 
-    if (missing(x))
-        return(.new_empty_SVT_SparseArray(type))
+    if (is.null(dim)) {
+        if (missing(x))
+            return(.new_empty_SVT_SparseArray(type))
+        return(.SVT_SparseArray(x, dimnames=dimnames, type=type))
+    }
 
-    .SVT_SparseArray(x, type=type)
+    dim <- S4Arrays:::normarg_dim(dim)
+    ans <- new_SVT_SparseArray(dim, dimnames=dimnames, check=FALSE)
+    if (!missing(x)) {
+        nzidx <- nzwhich(x)
+        ans[nzidx] <- as.vector(x[nzidx])
+    }
+    if (!identical(type, NA))
+        type(ans) <- type
+    ans
 }
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Coercion to SparseArray, and the SparseArray() constructor
+### The SparseArray() constructor
 ###
-### The SVT_SparseArray representation is preferred over COO_SparseArray when
-### coercing to SparseArray or when using the SparseArray() constructor. With
-### one exception: when the object to coerce is an RsparseMatrix derivative!
-### This is because we don't have an efficient way to coerce these objects to
-### SVT_SparseArray at the moment.
-
-setAs("ANY", "SparseArray", function(from) as(from, "SVT_SparseArray"))
-setAs("ANY", "SparseMatrix", function(from) as(from, "SVT_SparseMatrix"))
-setAs("RsparseMatrix", "SparseArray",
-    function(from) as(from, "COO_SparseArray")
-)
-setAs("RsparseMatrix", "SparseMatrix",
-    function(from) as(from, "COO_SparseMatrix")
-)
 
+### Preference is given to the SVT_SparseArray representation.
 SparseArray <- function(x, type=NA)
 {
     if (!identical(type, NA))
@@ -398,14 +562,18 @@ SparseArray <- function(x, type=NA)
         return(.new_empty_SVT_SparseArray(type))
 
     if (is(x, "SparseArray")) {
-        check_svt_version(x)
+        if (is(x, "SVT_SparseArray"))
+            check_svt_version(x)
         if (!identical(type, NA))
             type(x) <- type
         return(x)
     }
 
+    ## Calling SparseArray() on a sparseMatrix derivative (e.g. on a
+    ## CsparseMatrix or TsparseMatrix object) produces an SVT_SparseMatrix
+    ## object. RsparseMatrix is the exception.
     if (is(x, "RsparseMatrix")) {
-        ans <- as(x, "COO_SparseArray")
+        ans <- as(x, "COO_SparseMatrix")
         if (!identical(type, NA))
             type(ans) <- type
         return(ans)
